include: add <atomic> header
[c11tester.git] / threads.h
index 248d948fe98c87863741d4ed94343ee22d9cef5e..91db236f4bddb904fbd7944c35c00b39bbfce092 100644 (file)
--- a/threads.h
+++ b/threads.h
 
 #include "mymemory.h"
 #include "libthreads.h"
-
-typedef int thread_id_t;
-
-#define THREAD_ID_T_NONE       -1
+#include "modeltypes.h"
 
 /** @brief Represents the state of a user Thread */
 typedef enum thread_state {
@@ -38,6 +35,7 @@ class ModelAction;
 /** @brief A Thread is created for each user-space thread */
 class Thread {
 public:
+       Thread(thread_id_t tid);
        Thread(thrd_t *t, void (*func)(void *), void *a);
        ~Thread();
        void complete();
@@ -79,29 +77,47 @@ public:
        bool wait_list_empty() { return wait_list.empty(); }
 
        /**
* Add a thread to the waiting list for this thread.
* @param t The Thread to add
         */
-       void push_wait_list(Thread *t) { wait_list.push_back(t); }
+       void push_wait_list(ModelAction *act) { wait_list.push_back(act); }
+
+       unsigned int num_wait_list() {
+               return wait_list.size();
+       }
+
+       ModelAction * get_waiter(unsigned int i) {
+               return wait_list[i];
+       }
 
+       ModelAction * get_pending() { return pending; }
+       void set_pending(ModelAction *act) { pending = act; }
        /**
* Remove one Thread from the waiting list
* @return The Thread that was removed from the waiting list
         */
-       Thread * pop_wait_list() {
-               Thread *ret = wait_list.front();
+       ModelAction * pop_wait_list() {
+               ModelAction *ret = wait_list.front();
                wait_list.pop_back();
                return ret;
        }
 
+       bool is_model_thread() { return model_thread; }
+
        friend void thread_startup();
 
-       SNAPSHOTALLOC
+       /**
+        * Intentionally NOT allocated with MODELALLOC or SNAPSHOTALLOC.
+        * Threads should be allocated on the user's normal (snapshotting) heap
+        * to allow their allocation/deallocation to follow the same pattern as
+        * the rest of the backtracked/replayed program.
+        */
 private:
        int create_context();
        Thread *parent;
        ModelAction *creation;
 
+       ModelAction *pending;
        void (*start_routine)(void *);
        void *arg;
        ucontext_t context;
@@ -111,11 +127,11 @@ private:
        thread_state state;
 
        /**
* A list of Threads waiting on this Thread. Particularly, this list is
* used for thread joins, where another Thread waits for this Thread to
* complete
         */
-       std::vector<Thread *> wait_list;
+       std::vector< ModelAction *, SnapshotAlloc<ModelAction *> > wait_list;
 
        /**
         * The value returned by the last action in this thread
@@ -123,6 +139,9 @@ private:
         * @see Thread::get_return_value()
         */
        uint64_t last_action_val;
+
+       /** @brief Is this Thread a special model-checker thread? */
+       const bool model_thread;
 };
 
 Thread * thread_current();
@@ -79,29 +77,47 @@ public:
 
        /**
-        * Add a thread to the waiting list for this thread.
-        * @param t The Thread to add
+        * Add a ModelAction to the waiting list for this thread.
+        * @param t The ModelAction to add. Must be a JOIN.
         */
@@ -79,29 +77,47 @@ public:
        /**
-        * Remove one Thread from the waiting list
-        * @return The Thread that was removed from the waiting list
+        * Remove one ModelAction from the waiting list
+        * @return The ModelAction that was removed from the waiting list
         */
@@ -111,11 +127,11 @@ private:
 
        /**
-        * A list of Threads waiting on this Thread. Particularly, this list is
-        * used for thread joins, where another Thread waits for this Thread to
-        * complete
+        * A list of ModelActions waiting on this Thread. Particularly, this
+        * list is used for thread joins, where another Thread waits for this
+        * Thread to complete
         */